<?php

namespace Drupal\k8s\Controller;

use Drupal\Component\Serialization\Yaml;
use Drupal\Core\Url;
use Drupal\Core\Access\AccessResult;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Routing\RouteProviderInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\cloud\Entity\CloudProjectInterface;
use Drupal\k8s\Entity\K8sNodeInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\Routing\Route;

/**
 * Controller responsible for K8s Node Allocated Resources Chart.
 */
class K8sNodeAllocatedResourcesChartController extends ControllerBase {

  /**
   * The route provider.
   *
   * @var \Drupal\Core\Routing\RouteProviderInterface
   */
  protected $routeProvider;

  /**
   * The current request.
   *
   * @var \Symfony\Component\HttpFoundation\Request
   */
  protected $request;

  /**
   * The database service.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $database;

  /**
   * Drupal\Core\Entity\EntityTypeManagerInterface definition.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * K8sNodeAllocatedResourcesChartController constructor.
   *
   * @param \Drupal\Core\Routing\RouteProviderInterface $route_provider
   *   Route Provider.
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   Request.
   * @param \Drupal\Core\Database\Connection $database
   *   The database service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   */
  public function __construct(RouteProviderInterface $route_provider, Request $request, Connection $database, EntityTypeManagerInterface $entity_type_manager) {
    $this->routeProvider = $route_provider;
    $this->request = $request;
    $this->database = $database;
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * Dependency Injection.
   *
   * @param \Symfony\Component\DependencyInjection\ContainerInterface $container
   *   The Instance of ContainerInterface.
   *
   * @return K8sNamespaceCostsChartController
   *   return created object.
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('router.route_provider'),
      $container->get('request_stack')->getCurrentRequest(),
      $container->get('database'),
      $container->get('entity_type.manager')
    );
  }

  /**
   * Checks user access for k8s node allocated resources chart.
   *
   * @param \Drupal\Core\Session\AccountInterface $account
   *   Run access checks for this account.
   * @param \Symfony\Component\Routing\Route $route
   *   The route object.
   *
   * @return \Drupal\Core\Access\AccessResult
   *   The access result.
   */
  public function access(AccountInterface $account, Route $route) {
    global $base_url;
    // Get the referer url.
    $referer = $this->request->headers->get('referer');
    if (!empty($referer)) {
      // Get the alias or the referer.
      $alias = substr($referer, strlen($base_url));
      $url = Url::fromUri("internal:$alias");
      if ($url->access($account)) {
        return AccessResult::allowed();
      }
    }
    return AccessResult::forbidden();
  }

  /**
   * Get allocated resources for all node.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   A JSON response of a K8s Node Allocated Resources.
   */
  public function getAllK8sNodeAllocatedResources() {
    $node_ids = $this->entityTypeManager()
      ->getStorage('k8s_node')
      ->getQuery()
      ->execute();
    $nodes = $this->entityTypeManager()
      ->getStorage('k8s_node')->loadMultiple($node_ids);

    return $this->calculateNodeAllocatedResources($nodes);
  }

  /**
   * Get K8s Node Allocated Resources.
   *
   * @param string $cloud_context
   *   Cloud context string.
   * @param \Drupal\cloud\Entity\CloudProjectInterface $cloud_project
   *   The cloud project.
   * @param \Drupal\k8s\Entity\K8sNodeInterface $k8s_node
   *   K8s Node entity.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   A JSON response of a K8s Node Allocated Resources.
   */
  public function getK8sNodeAllocatedResources($cloud_context, CloudProjectInterface $cloud_project = NULL, K8sNodeInterface $k8s_node = NULL) {
    $nodes = [];
    $cloud_contexts = [];

    if (!empty($cloud_project)) {
      $k8s_clusters = $cloud_project->get('field_k8s_clusters')->getValue();
      foreach ($k8s_clusters ?: [] as $k8s_cluster) {
        $cloud_contexts[] = $k8s_cluster['value'];
      }
    }
    elseif (!empty($cloud_context)) {
      $cloud_contexts[] = $cloud_context;
    }

    // If cloud_context is passed, filter entities with it.
    if (!empty($cloud_contexts)) {
      foreach ($cloud_contexts ?: [] as $cloud_context) {
        $nodes = array_merge($nodes, isset($k8s_node)
          ? $this->entityTypeManager()->getStorage('k8s_node')
            ->loadByProperties([
              'id' => $k8s_node->id(),
            ])
          : $this->entityTypeManager()->getStorage('k8s_node')
            ->loadByProperties([
              'cloud_context' => $cloud_context,
            ]));
      }
    }
    return !empty($cloud_project)
      ? $this->calculateNodeAllocatedResources($nodes)
      : $this->calculateNodeAllocatedResources($nodes, $cloud_context);
  }

  /**
   * Calculate node allocated resources.
   *
   * @param array $nodes
   *   Array of K8s Node entities.
   * @param string $cloud_context
   *   Optional cloud context.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   Json response.
   */
  private function calculateNodeAllocatedResources(array $nodes, $cloud_context = NULL) {

    $period = $this->request->get('period');

    if (!isset($period) || empty($period)) {
      $period = 14;
    }

    $node_metrics = $this->getAllNodesMetrics($cloud_context, $period);
    $response = [];

    foreach ($node_metrics ?: [] as $time => $node_datas) {
      $cpu_usage = (float) 0;
      $memory_usage = (float) 0;
      $pod_usage = 0;
      $cpu_capacity = (float) 0;
      $memory_capacity = (float) 0;
      $pod_capacity = 0;
      $data = [];
      foreach ($nodes ?: [] as $node) {
        $key = $node->getName();
        if (isset($node_datas[$key])) {
          $node_data = $node_datas[$key];
          $cpu_usage += $node_data['cpu'];
          $memory_usage += $node_data['memory'];
          $pod_usage += $node_data['pod_allocation'];
          $cpu_capacity += $node_data['cpu_capacity'];
          $memory_capacity += $node_data['memory_capacity'];
          $pod_capacity += $node_data['pod_capacity'];
        }
      }
      $data['cpu'] = [
        'timestamp' => $time,
        'cpu_usage' => $cpu_usage,
        'cpu_capacity' => $cpu_capacity,
      ];
      $data['memory'] = [
        'timestamp' => $time,
        'memory_usage' => $memory_usage,
        'memory_capacity' => $memory_capacity,
      ];
      $data['pod'] = [
        'timestamp' => $time,
        'pod_usage' => $pod_usage,
        'pod_capacity' => $pod_capacity,
      ];
      $response[] = $data;
    }

    return new JsonResponse($response);
  }

  /**
   * Get all nodes metrics.
   *
   * @param string $cloud_context
   *   Cloud context string.  If no cloud context passed, retrieve
   *   all metric messages.
   * @param int $period
   *   The period to get data.
   *
   * @return array
   *   The node metrics array.
   */
  private function getAllNodesMetrics($cloud_context = NULL, $period = 14) {
    $query = $this->database
      ->select('watchdog', 'w')
      ->fields('w', ['message', 'timestamp'])
      ->condition('type', 'k8s_metrics')
      ->condition('timestamp', time() - (int) $period * 24 * 60 * 60, '>=');

    // Add cloud_context if set.
    if (!empty($cloud_context)) {
      $query->condition('message', "%$cloud_context:%", 'LIKE');
    }

    $result = $query->condition('message', '%nodes:%', 'LIKE')
      ->orderBy('timestamp')
      ->execute();
    $node_metrics = [];
    foreach ($result ?: [] as $record) {
      $message = Yaml::decode($record->message);
      if (!empty($cloud_context)) {
        if (empty($message[$cloud_context])) {
          continue;
        }
        if (empty($message[$cloud_context]['nodes'])) {
          continue;
        }
        $nodes = $message[$cloud_context];
      }
      else {
        $nodes = array_shift($message);
      }
      $date = date('Y-m-d H:i', $record->timestamp);
      $timestamp = strtotime($date);
      if (!isset($node_metrics[$timestamp])) {
        $node_metrics[$timestamp] = [];
      }
      foreach ($nodes['nodes'] ?: [] as $key => $node_data) {
        if (isset($node_data['cpu_capacity']) && isset($node_data['memory_capacity']) && isset($node_data['pod_capacity'])) {
          $node_metrics[$timestamp][$key] = [
            'cpu' => $node_data['cpu'],
            'cpu_capacity' => $node_data['cpu_capacity'],
            'memory' => $node_data['memory'] ?? 0,
            'memory_capacity' => $node_data['memory_capacity'],
            'pod_allocation' => $node_data['pod_allocation'],
            'pod_capacity' => $node_data['pod_capacity'],
          ];
        }
      }
    }

    return $node_metrics;
  }

  /**
   * Get a list of chart period.
   *
   * @param string $cloud_context
   *   The cloud context.
   * @param \Drupal\cloud\Entity\CloudProjectInterface $cloud_project
   *   The cloud project.
   * @param string $json
   *   Whether the response should be json.
   *
   * @return array|\Symfony\Component\HttpFoundation\JsonResponse
   *   The list or JSON response of chart period.
   */
  public function getEc2ChartPeriod($cloud_context = NULL, CloudProjectInterface $cloud_project = NULL, $json = NULL) {
    $cloud_contexts = [];

    if (!empty($cloud_project)) {
      $k8s_clusters = $cloud_project->get('field_k8s_clusters')->getValue();
      foreach ($k8s_clusters ?: [] as $k8s_cluster) {
        $cloud_contexts[] = $k8s_cluster['value'];
      }
    }
    elseif (!empty($cloud_context)) {
      $cloud_contexts[] = $cloud_context;
    }

    try {
      $query = $this->database->select('watchdog', 'w');
      $query->addExpression('min(w.timestamp)');
      if (!empty($cloud_contexts)) {
        $or_group = $query->orConditionGroup();
        foreach ($cloud_contexts ?: [] as $cloud_context) {
          $or_group->condition('message', "%$cloud_context%", 'LIKE');
        }
      }
      $timestamp = $query->condition('type', 'k8s_metrics')
        ->execute()->fetchField();
    }
    catch (\Exception $e) {
      $this->handleException($e);
    }

    $period_option = [
      1 => $this->t('One day'),
      7 => $this->t('One week'),
      14 => $this->t('Two weeks'),
      31 => $this->t('One month'),
      183 => $this->t('Half a year'),
      365 => $this->t('One year'),
    ];
    $keys = array_keys($period_option);
    $keys = array_reverse($keys);
    foreach ($keys ?: [] as $idx => $day) {
      if ($day > 7 && (time() - $day * 24 * 60 * 60 <= $timestamp) && (time() - $keys[$idx + 1] * 24 * 60 * 60 <= $timestamp)) {
        unset($period_option[$day]);
      }
    }
    if (isset($json)) {
      return new JsonResponse($period_option);
    }
    else {
      return $period_option;
    }
  }

}
